from fractions import Fraction
from pathlib import Path

from more_itertools import first_true

from ._aliases import AudioStream, IndexedAudioStream, IndexedOutputs, Millisecond, StreamIndex


class FFmpegJob:
    def __init__(
        self,
        input_file: Path,
        audio_streams: list[AudioStream],
        indexed_outputs: IndexedOutputs,
    ):
        self.input_file: Path = input_file
        self.indexed_outputs: IndexedOutputs = indexed_outputs
        self.audio_streams = audio_streams

        indexed_audio_streams = {}
        for stream in audio_streams:
            indexed_audio_streams[stream["index"]] = stream
        self.indexed_audio_streams: IndexedAudioStream = indexed_audio_streams

        # TODO: Do we need this?
        self.durations_ms_dict: dict[StreamIndex, Millisecond] = {
            stream["index"]: self.stream_duration_ms(stream) for stream in audio_streams
        }

    @classmethod
    def stream_duration_ms(cls, stream: AudioStream) -> Millisecond:
        return 1_000_000 * float(stream["duration_ts"]) * float(Fraction(stream["time_base"]))

    def stream(self, index: StreamIndex) -> AudioStream:
        stream: AudioStream | None = first_true(
            self.audio_streams,
            default=None,
            pred=lambda st: st["index"] == index,
        )

        if stream is None:
            raise IndexError(f'Stream with index "{index}" not found')

        return stream



import inspect
import logging
from logging import FileHandler, Logger
from pathlib import Path

from rich.logging import RichHandler

from BAET._console import app_console

rich_handler = RichHandler(rich_tracebacks=True, console=app_console)
logging.basicConfig(
    level=logging.INFO,
    format="%(threadName)s: %(message)s",
    datefmt="[%X]",
    handlers=[rich_handler],
)

app_logger = logging.getLogger("app_logger")


def create_logger() -> Logger:
    frame = inspect.stack()[1]
    module = inspect.getmodule(frame[0])

    if module is None:
        raise RuntimeError("Could not inspect module")

    module_name = module.__name__

    logger = app_logger.getChild(module_name)
    return logger


def configure_logging(*, enable_logging: bool = True, file_out: Path | None = None) -> None:
    if not enable_logging:
        logging.disable(logging.CRITICAL)

    if file_out is not None:
        handler = FileHandler(filename=file_out)
        handler.setFormatter(rich_handler.formatter)
        app_logger.addHandler(handler)



from collections.abc import Mapping
from typing import Any, TypeAlias

from bidict import BidirectionalMapping
from ffmpeg import Stream
from rich.progress import TaskID

# Numbers
Millisecond: TypeAlias = int | float

# FFmpeg
StreamIndex: TypeAlias = int
AudioStream: TypeAlias = dict[str, Any]

# Mappings
IndexedOutputs: TypeAlias = Mapping[StreamIndex, Stream]
IndexedAudioStream: TypeAlias = Mapping[StreamIndex, AudioStream]
StreamTaskBiMap: TypeAlias = BidirectionalMapping[StreamIndex, TaskID]



from __future__ import annotations

from enum import StrEnum
from typing import Literal, TypeAlias

from rich.console import Console, ConsoleOptions, RenderResult
from rich.repr import Result

ProgressStatusLiteral: TypeAlias = Literal["Waiting", "Running", "Completed", "Error"]


class ProgressStatus(StrEnum):
    Waiting = "Waiting"
    Running = "Running"
    Completed = "Completed"
    Error = "Error"

    def __rich_repr__(self) -> Result:
        yield repr(self)

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        yield repr(self)


ProgressStatusType: TypeAlias = ProgressStatus | ProgressStatusLiteral

if __name__ == "__main__":
    import rich

    waiting = ProgressStatus("Waiting")
    completed = ProgressStatus(ProgressStatus.Completed)

    rich.print(waiting)
    rich.print(completed)



import contextlib
from collections.abc import Generator, Iterator, MutableMapping
from pathlib import Path
from typing import Any

import ffmpeg
from ffmpeg import Stream
from rich.console import Console, ConsoleOptions, Group, RenderResult
from rich.padding import Padding
from rich.prompt import Confirm
from rich.table import Table

from ._aliases import AudioStream
from ._console import app_console, error_console
from ._constants import VIDEO_EXTENSIONS
from ._logging import create_logger
from .app_args import AppArgs, InputFilters, OutputConfigurationOptions
from .job_progress import FFmpegJobProgress
from .jobs import FFmpegJob

logger = create_logger()


@contextlib.contextmanager
def probe_audio_streams(file: Path) -> Iterator[list[AudioStream]]:
    try:
        logger.info('Probing file "%s"', file)
        probe = ffmpeg.probe(file)

        audio_streams = sorted(
            [stream for stream in probe["streams"] if "codec_type" in stream and stream["codec_type"] == "audio"],
            key=lambda stream: stream["index"],
        )

        if not audio_streams:
            logger.warning("No audio streams found")
            yield []
            return

        logger.info("Found %d audio streams", len(audio_streams))
        yield audio_streams

    except (ffmpeg.Error, ValueError) as e:
        logger.critical("%s: %s", type(e).__name__, e)
        error_console.print_exception()
        raise e


def can_write_file(file: Path, has_overwrite_permission: bool) -> bool:
    if not file.exists():
        return True

    if has_overwrite_permission:
        return True

    return Confirm.ask(  # type: ignore
        f'The file "{file.name}" already exists. Overwrite?',
        console=app_console,
    )


class FileSourceDirectory:
    def __init__(self, directory: Path, filters: InputFilters):
        if not directory.is_dir():
            raise NotADirectoryError(directory)  # FIXME: Is this right?

        self._directory = directory.resolve(strict=True)
        self._filters = filters

    def __iter__(self) -> Generator[Path, Any, None]:
        for file in self._directory.iterdir():
            if not file.is_file():
                continue

            # TODO: This should be moved.
            #  Rather than checking a global variable, it should be provided somehow.
            #  Perhaps via command line args.
            if file.suffix not in VIDEO_EXTENSIONS:
                continue

            if self._filters.include.match(file.name):
                if self._filters.exclude is None:
                    yield file
                elif self._filters.exclude.match(file.name):
                    yield file


class MultitrackAudioBulkExtractorJobs:
    def __init__(
        self,
        input_dir: Path,
        output_dir: Path,
        filters: InputFilters,
        output_configuration: OutputConfigurationOptions,
    ):
        self._input_dir = input_dir
        self._output_dir = output_dir
        self._filters = filters
        self._output_configuration = output_configuration
        self._file_source_directory = FileSourceDirectory(input_dir, filters)

    def _create_output_filepath(self, file: Path, stream_index: int) -> Path:
        filename = Path(f"{file.stem}_track{stream_index}.{self._output_configuration.file_type}")

        out_path = self._output_dir if self._output_configuration.no_output_subdirs else self._output_dir / file.stem

        out_path.mkdir(parents=True, exist_ok=True)
        return out_path / filename

    def build_job(self, file: Path) -> FFmpegJob:
        logger.info(f'Building job for "{file}"')

        audio_streams: list[AudioStream] = []
        indexed_outputs: MutableMapping[int, Stream] = {}

        file = file.expanduser()
        with probe_audio_streams(file) as streams:
            for idx, stream in enumerate(streams):
                ffmpeg_input = ffmpeg.input(str(file))
                stream_index = stream["index"]
                output_path = self._create_output_filepath(file, stream_index)
                sample_rate = stream.get(
                    "sample_rate",
                    self._output_configuration.fallback_sample_rate,
                )

                can_write = can_write_file(
                    file=output_path,
                    has_overwrite_permission=self._output_configuration.overwrite_existing,
                )

                if can_write:
                    # Add stream here since otherwise there will possibly be more streams to indexes
                    # TODO: Maybe make a function/class to help with this?
                    audio_streams.append(stream)

                    indexed_outputs[stream_index] = (
                        ffmpeg.output(
                            ffmpeg_input[f"a:{idx}"],
                            str(output_path),
                            acodec=self._output_configuration.acodec,
                            audio_bitrate=sample_rate,
                        )
                        .overwrite_output()
                        .global_args("-progress", "-", "-nostats")
                    )

        return FFmpegJob(file, audio_streams, indexed_outputs)

    def __iter__(self) -> Generator[FFmpegJob, Any, None]:
        yield from map(self.build_job, self._file_source_directory)


class MultiTrackAudioBulkExtractor:
    def __init__(self, app_args: AppArgs) -> None:
        logger.info("Starting MultiTrackAudioBulkExtractor")

        self._app_args = app_args
        self._extractor_jobs = MultitrackAudioBulkExtractorJobs(
            app_args.input_dir,
            app_args.output_dir,
            app_args.input_filters,
            app_args.output_configuration,
        )

        # Need to compile the jobs here, otherwise rich bugs out when prompting.
        # This is because run_synchronously is called within a live display, which is buggy.
        # In the future, a refactor to separate the running of jobs from this class can make things less complex
        self._jobs = list(self._extractor_jobs)

        self.display = Table.grid()

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        yield self.display

    def run_synchronously(self) -> None:
        job_progresses = [FFmpegJobProgress(job) for job in self._jobs]
        self.display.add_row(Padding(Group(*job_progresses), pad=(1, 2)))

        logger.info("Starting synchronous execution of queued jobs")
        for progress in job_progresses:
            logger.info(f'Starting job "{progress.job.input_file}"')
            progress.start()



import subprocess
from os import PathLike

from ._console import error_console
from ._logging import create_logger

logger = create_logger()


def which_ffmpeg() -> str | PathLike[str] | None:
    from shutil import which

    return which("ffmpeg")


def get_ffmpeg_version() -> str | None:
    try:
        ffmpeg = which_ffmpeg()

        if not ffmpeg:
            return None

        proc = subprocess.run([ffmpeg, "-version"], capture_output=True)

        if proc.returncode != 0:
            err = proc.stderr.decode("utf-8")
            raise RuntimeError(f"FFmpeg returned non-zero exit code when getting version:\n{err}")

        output = proc.stdout.decode("utf-8")
        return output[14 : output.find("Copyright")].strip()

    except RuntimeError as e:
        logger.critical("%s: %s", type(e).__name__, e)
        error_console.print_exception()
        raise e


class FFmpegVersionInfo:
    def __init__(self) -> None:
        self._version: str | None = None

    @property
    def version(self) -> str:
        if not self._version:
            self._version = get_ffmpeg_version()

        return self._version or "None"

    def __str__(self) -> str:
        return self.version


ffmpeg_version_info = FFmpegVersionInfo()

if __name__ == "__main__":
    version_info = FFmpegVersionInfo()
    print("Version:", version_info)



from collections.abc import Generator, Mapping
from typing import Any

from rich.console import Console, ConsoleOptions, RenderResult
from rich.repr import Result
from rich.style import Style
from rich.text import Text

from BAET._theme import app_theme
from BAET.progress_status import ProgressStatus, ProgressStatusLiteral, ProgressStatusType


def parse_style(style: Style | str) -> Style:
    if isinstance(style, Style):
        return style

    return Style.parse(style)


class ProgressStyle:
    def __init__(
        self,
        style_dict: dict[ProgressStatusType, str | Style] | None = None,
        *,
        waiting_style: str | Style | None = None,
        running_style: str | Style | None = None,
        completed_style: str | Style | None = None,
        error_style: str | Style | None = None,
    ):
        default_style_dict: dict[ProgressStatus, Style] = {
            key: parse_style(value or default)
            for key, value, default in [
                (ProgressStatus.Waiting, waiting_style, "dim bright_white"),
                (ProgressStatus.Running, running_style, "bright_cyan"),
                (ProgressStatus.Completed, completed_style, "bright_green"),
                (ProgressStatus.Error, error_style, "bright_red"),
            ]
        }

        if style_dict is not None:
            parsed_styles = {ProgressStatus(key): parse_style(val) for key, val in style_dict.items()}

            default_style_dict.update(parsed_styles)

        self.style_dict: Mapping[ProgressStatus, Style] = default_style_dict

    def __call__(
        self,
        text: str,
        status: ProgressStatus | ProgressStatusLiteral,
        *args: Any,
        style: Style | None = None,
        **kwargs: Any,
    ) -> Text:
        return Text(
            text,
            self.style_dict[ProgressStatus(status)] + style,
            *args,
            **kwargs,
        )

    def __rich_repr__(self) -> Result:
        yield "style_dict", self.style_dict

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        with console.capture() as capture:
            console.print(self.style_dict)
        yield Text.from_ansi(capture.get())


if __name__ == "__main__":
    from rich.console import Group, group
    from rich.padding import Padding
    from rich.panel import Panel
    from rich.pretty import pprint

    console = Console(theme=app_theme)

    basic_style = ProgressStyle(
        {"Completed": "bold green", ProgressStatus.Error: "bold red"},
        running_style="italic bold cyan",
        waiting_style="underline yellow",
    )

    theme_style = ProgressStyle(
        {
            ProgressStatus.Waiting: app_theme.styles["status.waiting"],
            ProgressStatus.Running: app_theme.styles["status.running"],
            ProgressStatus.Completed: app_theme.styles["status.completed"],
            ProgressStatus.Error: app_theme.styles["status.error"],
        },
    )

    @group()
    def apply_styles(style: ProgressStyle, message: str) -> Generator[Text, Any, None]:
        for status in ["Waiting", "Completed", "Running", "Error"]:
            yield style(message.format(status), status)  # type: ignore

    @group()
    def make_style_group(style: ProgressStyle, message: str) -> Generator[Text, Any, None]:
        with console.capture() as pretty_capture:
            pprint(
                {key.value: val.color.name for key, val in style.style_dict.items() if val.color},
                console=console,
                expand_all=True,
                indent_guides=False,
            )
        with console.capture() as repr_capture:
            console.print(
                style,
                "\n[u]Demo Application:[/]",
                Padding(apply_styles(style, message), pad=(0, 0, 0, 3)),
            )

        yield Text.from_ansi(f"Input Style: {pretty_capture.get()}\n")
        yield Text.from_ansi(f"Resulting ProgressStyle: {repr_capture.get()}")

    layout = Group(
        Panel(
            make_style_group(basic_style, "This is the style for ProgressStyle.{0}"),
            title="[bold green]Basic Style",
        ),
        Panel(
            make_style_group(
                theme_style,
                'This is the style for "ProgressStyle.{0}", taken from app_theme in BAET',
            ),
            title="[bold green]Theme Style",
        ),
    )

    console.print(layout)



from rich.console import Console
from rich.progress import Progress, TextColumn

from .progress_status import ProgressStatus
from .progress_style import ProgressStyle


class ProgressCheckList:
    def __init__(
        self,
        waiting_description: str,
        running_description: str,
        completed_description: str,
        error_description: str,
        progress_style: ProgressStyle | None = None,
        console: Console | None = None,
    ):
        self.descriptions: dict[ProgressStatus, str] = {
            ProgressStatus.Waiting: waiting_description,
            ProgressStatus.Running: running_description,
            ProgressStatus.Completed: completed_description,
            ProgressStatus.Error: error_description,
        }

        self.progress_style = progress_style or ProgressStyle()

        self.overall_progress = Progress(
            TextColumn("{task.description}"),
            console=console,
        )

        self.overall_progress_task = self.overall_progress.add_task(
            waiting_description,
            total=None,
            start=False,
        )

    def add_item(
        self,
        waiting_description: str,
        running_description: str,
        completed_description: str,
        error_description: str,
    ) -> None:
        # TODO
        pass



import argparse
import re
from argparse import ArgumentParser
from pathlib import Path
from re import Pattern

from pydantic import BaseModel, ConfigDict, DirectoryPath, Field, field_validator
from rich.console import Console, ConsoleOptions, RenderResult
from rich.markdown import Markdown
from rich.padding import Padding
from rich.table import Table
from rich.terminal_theme import DIMMED_MONOKAI
from rich.text import Text
from rich_argparse import HelpPreviewAction, RichHelpFormatter
from typing_extensions import Annotated

from ._console import app_console
from ._metadata import app_version
from ._path import ffmpeg_version_info

file_type_pattern = re.compile(r"^\.?(\w+)$")


class InputFilters(BaseModel):
    include: Pattern[str] = Field(...)
    exclude: Pattern[str] | None = Field(...)

    @field_validator("include", mode="before")
    @classmethod
    def validate_include_nonempty(cls, v: str) -> Pattern[str]:
        if v is None or not v.strip():
            return re.compile(".*")
        return re.compile(v)

    @field_validator("exclude", mode="before")
    @classmethod
    def validate_exclude_nonempty(cls, v: str) -> Pattern[str] | None:
        if v is None or not v.strip():
            return None
        return re.compile(v)


class OutputConfigurationOptions(BaseModel):
    overwrite_existing: bool = Field(...)
    no_output_subdirs: bool = Field(...)
    acodec: str = Field(...)
    fallback_sample_rate: Annotated[int, Field(gt=0)] = Field(...)
    file_type: str = Field(...)

    @field_validator("file_type", mode="before")
    @classmethod
    def validate_file_type(cls, v: str) -> str:
        matched = file_type_pattern.match(v)
        if matched:
            return matched.group(1)
        raise ValueError(f"Invalid file type: {v}")


class DebugOptions(BaseModel):
    logging: bool = Field(...)
    dry_run: bool = Field(...)
    trim: Annotated[int, Field(gt=0)] | None = Field(...)
    print_args: bool = Field(...)
    show_ffmpeg_cmd: bool = Field(...)
    run_synchronously: bool = Field(...)


class AppDescription:
    @staticmethod
    def __rich_console__(console: Console, options: ConsoleOptions) -> RenderResult:
        yield Padding(Markdown("# Bulk Audio Extract Tool"), pad=(1, 0))
        yield "Extract audio from a directory of videos using FFMPEG.\n"

        website_link = "https://github.com/TimeTravelPenguin/BulkAudioExtractTool"
        desc_kvps = [
            (
                Text("App name:", justify="right"),
                Text(
                    "Bulk Audio Extract Tool (BAET)",
                    style="argparse.prog",
                    justify="left",
                ),
            ),
            (
                Text("App Version:", justify="right"),
                Text(app_version(), style="app.version", justify="left"),
            ),
            (
                Text("FFmpeg Version:", justify="right"),
                Text(ffmpeg_version_info.version, style="app.version", justify="left"),
            ),
            (
                Text("Author:", justify="right"),
                Text("Phillip Smith", style="bright_yellow", justify="left"),
            ),
            (
                Text("Website:", justify="right"),
                Text(
                    website_link,
                    style=f"underline blue link {website_link}",
                    justify="left",
                ),
            ),
        ]

        grid = Table.grid(expand=True)
        grid.add_column(justify="left")
        grid.add_column(justify="right")
        for key, value in desc_kvps:
            grid.add_row(Padding(key, (0, 3, 0, 0)), value)

        yield grid


def new_empty_argparser() -> ArgumentParser:
    def get_formatter(prog: str) -> RichHelpFormatter:
        return RichHelpFormatter(prog, max_help_position=40, console=app_console)

    # todo: use console protocol https://rich.readthedocs.io/en/stable/protocol.html#console-protocol
    description = AppDescription()

    RichHelpFormatter.highlights.append(
        r"(?P<arg_default_parens>\((?P<arg_default>Default: (?P<arg_default_value>.*))\))"
    )

    RichHelpFormatter.highlights.append(r"(?P<help_keyword>ffmpeg|ffprobe)")
    RichHelpFormatter.highlights.append(r"(?P<debug_todo>\[TODO\])")

    return argparse.ArgumentParser(
        prog="Bulk Audio Extract Tool (src)",
        description=description,  # type: ignore
        epilog=Markdown(
            "Phillip Smith, 2024",
            justify="right",
            style="argparse.prog",
        ),
        formatter_class=get_formatter,
    )


class AppArgs(BaseModel):
    """Application commandline arguments.

    Raises:
        ValueError: The provided path is not a directory.

    Returns:
        DirectoryPath: Validated directory path.
    """

    model_config = ConfigDict(frozen=True, from_attributes=True)

    input_dir: DirectoryPath = Field(...)
    output_dir: DirectoryPath = Field(...)
    input_filters: InputFilters = Field(...)
    output_configuration: OutputConfigurationOptions = Field(...)
    debug_options: DebugOptions = Field(...)


def get_args() -> AppArgs:
    parser = new_empty_argparser()

    parser.add_argument(
        "--version",
        action="version",
        version=f"[argparse.prog]%(prog)s[/] version [i]{app_version()}[/]",
    )

    io_group = parser.add_argument_group(
        "Input/Output",
        "Options to control the source and destination directories of input and output files.",
    )

    io_group.add_argument(
        "-i",
        "--input-dir",
        action="store",
        type=DirectoryPath,
        metavar="INPUT_DIR",
        required=True,
        help="Source directory.",
    )

    io_group.add_argument(
        "-o",
        "--output-dir",
        default=None,
        action="store",
        type=Path,
        help="Destination directory. Default is set to the input directory. To use the current directory, "
        'use [blue]"."[/]. (Default: None)',
    )

    query_group = parser.add_argument_group(
        title="Input Filter Configuration",
        description="Configure how the application includes and excludes files to process.",
    )

    query_group.add_argument(
        "--include",
        default=None,
        metavar="REGEX",
        help='[TODO] If provided, only include files that match a regex pattern. (Default: ".*")',
    )

    query_group.add_argument(
        "--exclude",
        default=None,
        metavar="REGEX",
        help="[TODO] If provided, exclude files that match a regex pattern. (Default: None)",
    )

    output_group = parser.add_argument_group(
        title="Output Configuration",
        description="Override the default output behavior of the application.",
    )

    output_group.add_argument(
        "--overwrite-existing",
        "--overwrite",
        default=False,
        action="store_true",
        help="Overwrite a file if it already exists. (Default: False)",
    )

    output_group.add_argument(
        "--no-output-subdirs",
        default=False,
        action="store_true",
        help="Do not create subdirectories for each video's extracted audio tracks in the output directory. "
        "(Default: True)",
    )

    output_group.add_argument(
        "--acodec",
        default="pcm_s16le",
        metavar="CODEC",
        help='[TODO] The audio codec to use when extracting audio. (Default: "pcm_s16le")',
    )

    output_group.add_argument(
        "--fallback-sample-rate",
        default=48000,
        metavar="RATE",
        help="[TODO] The sample rate to use if it cannot be determined via [blue]ffprobe[/]. (Default: 48000)",
    )

    output_group.add_argument(
        "--file-type",
        default="wav",
        metavar="EXT",
        help='[TODO] The file type to use for the extracted audio. (Default: "wav")',
    )

    debug_group = parser.add_argument_group(
        "Debugging",
        "Options to help debug the application.",
    )

    debug_group.add_argument(
        "--run-synchronously",
        "--sync",
        default=False,
        action="store_true",
        help="[TODO] Run each each job in order. This should reduce the CPU workload, but may increase runtime. A "
        "'job' is per file input, regardless of whether ffmpeg commands are merged (see: "
        "`--output-streams-separately`). (Default: False)",
    )

    debug_group.add_argument(
        "--logging",
        default=False,
        action="store_true",
        help="[TODO] Show the logging of application execution. (Default: False)",
    )

    debug_group.add_argument(
        "--print-args",
        default=False,
        action="store_true",
        help="Print the parsed arguments and exit. (Default: False)",
    )

    debug_group.add_argument(
        "--dry-run",
        default=False,
        action="store_true",
        help="Run the program without actually extracting any audio. (Default: False)",
    )

    debug_group.add_argument(
        "--show-ffmpeg-cmd",
        "--cmds",
        default=False,
        action="store_true",
        help="[TODO] Print to the console the generated ffmpeg command. (Default: False)",
    )

    debug_group.add_argument(
        "--trim-short",
        default=None,
        type=int,
        dest="trim",
        help="[TODO] Trim the audio to the specified number of seconds. This is useful for testing. (Default: None)",
    )

    parser.add_argument(
        "--generate-help-preview",
        action=HelpPreviewAction,
        path="help-preview.svg",  # (optional) or "help-preview.html" or "help-preview.txt"
        export_kwds={"theme": DIMMED_MONOKAI},  # (optional) keywords passed to console.save_... methods
        help=argparse.SUPPRESS,
    )

    args = parser.parse_args()

    input_filters = InputFilters(include=args.include, exclude=args.exclude)

    output_config = OutputConfigurationOptions(
        overwrite_existing=args.overwrite_existing,
        no_output_subdirs=args.no_output_subdirs,
        acodec=args.acodec,
        fallback_sample_rate=args.fallback_sample_rate,
        file_type=args.file_type,
    )

    debug_options = DebugOptions(
        logging=args.logging,
        dry_run=args.dry_run,
        trim=args.trim,
        print_args=args.print_args,
        show_ffmpeg_cmd=args.show_ffmpeg_cmd,
        run_synchronously=args.run_synchronously,
    )

    output_dir = args.output_dir or args.input_dir
    output_dir = output_dir.expanduser()
    output_dir.mkdir(parents=True, exist_ok=True)

    app_args: AppArgs = AppArgs.model_validate(
        {
            "input_dir": args.input_dir.expanduser(),
            "output_dir": output_dir,
            "input_filters": input_filters,
            "output_configuration": output_config,
            "debug_options": debug_options,
        },
        strict=True,
    )

    return app_args



import ffmpeg
from bidict import MutableBidirectionalMapping, bidict
from rich.console import Console, ConsoleOptions, ConsoleRenderable, Group, RenderResult
from rich.highlighter import ReprHighlighter
from rich.padding import Padding
from rich.progress import BarColumn, Progress, TaskID, TextColumn, TimeElapsedColumn, TimeRemainingColumn

from ._aliases import StreamTaskBiMap
from ._console import app_console
from ._logging import create_logger
from .jobs import FFmpegJob

logger = create_logger()


class FFmpegJobProgress(ConsoleRenderable):
    # TODO: Need mediator to consumer/producer printing
    def __init__(self, job: FFmpegJob):
        self.job = job

        bar_blue = "#5079AF"
        bar_yellow = "#CAAF39"

        self._overall_progress = Progress(
            TextColumn('Progress for "{task.fields[filename]}"', highlighter=ReprHighlighter()),
            BarColumn(
                complete_style=bar_blue,
                finished_style="green",
                pulse_style=bar_yellow,
            ),
            TextColumn("Completed {task.completed} of {task.total}", highlighter=ReprHighlighter()),
            TimeElapsedColumn(),
            TimeRemainingColumn(),
            console=app_console,
        )

        self._overall_progress_task = self._overall_progress.add_task(
            "Waiting...",
            start=False,
            filename=job.input_file.name,
            total=len(self.job.audio_streams),
        )

        self._stream_task_progress = Progress(
            TextColumn("Audio stream {task.fields[stream_index]}", highlighter=ReprHighlighter()),
            BarColumn(
                complete_style=bar_blue,
                finished_style="green",
                pulse_style=bar_yellow,
            ),
            TextColumn("{task.fields[status]}"),
            TimeElapsedColumn(),
            TimeRemainingColumn(),
            console=app_console,
        )

        self._stream_task_bimap: StreamTaskBiMap = bidict()

        stream_task_bimap: MutableBidirectionalMapping[int, TaskID] = bidict()
        for stream in self.job.audio_streams:
            stream_index = stream["index"]

            task = self._stream_task_progress.add_task(
                "Waiting...",
                start=False,
                total=self.job.stream_duration_ms(stream),
                stream_index=stream_index,
                status="[plum4]Waiting[/]",
            )

            stream_task_bimap[stream_index] = task

        self._stream_task_bimap = stream_task_bimap

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        yield Group(
            self._overall_progress,
            Padding(self._stream_task_progress, (1, 0, 1, 5)),
        )

    def _run_task(self, task: TaskID) -> None:
        stream_index: int = self._stream_task_bimap.inverse[task]

        logger.info(f"Extracting audio stream {stream_index} of {self.job.input_file.name}")

        output = self.job.indexed_outputs[stream_index]

        proc = ffmpeg.run_async(
            output,
            pipe_stdout=True,
            pipe_stderr=True,
        )

        try:
            with proc as p:
                for line in p.stdout:
                    decoded = line.decode("utf-8").strip()
                    if "out_time_ms" in decoded:
                        val = decoded.split("=", 1)[1]
                        self._stream_task_progress.update(
                            task,
                            completed=float(val),
                        )

            if p.returncode != 0:
                raise RuntimeError(p.stderr.read().decode("utf-8"))

        except RuntimeError as e:
            logger.critical("%s: %s", type(e).__name__, e)
            raise e

    def start(self) -> None:
        self._overall_progress.start_task(self._overall_progress_task)
        logger.info(f"Stream index to job task ID bimap: {self._stream_task_bimap}")
        for task in self._stream_task_bimap.values():
            self._stream_task_progress.start_task(task)

            self._stream_task_progress.update(task, status="[italic cornflower_blue]Working[/]")

            try:
                self._run_task(task)
                self._stream_task_progress.update(task, status="[bold green]Complete[/]")
            except RuntimeError as e:
                self._stream_task_progress.update(task, status="[bold red]ERROR[/]")
            finally:
                self._stream_task_progress.stop_task(task)
                self._overall_progress.advance(self._overall_progress_task, advance=1)

        self._overall_progress.stop_task(self._overall_progress_task)



from rich.theme import Theme

app_theme = Theme(
    {
        "app.version": "italic bright_cyan",
        "argparse.arg_default": "dim italic",
        "argparse.arg_default_parens": "dim",
        "argparse.arg_default_value": "not italic bold dim",
        "argparse.help_keyword": "bold blue",
        "argparse.debug_todo": "reverse bold indian_red",
        "status.waiting": "dim bright_white",
        "status.running": "bright_cyan",
        "status.completed": "bright_green",
        "status.error": "bright_red",
    }
)



import sys
from datetime import datetime
from pathlib import Path

import rich
from rich.live import Live
from rich.traceback import install

from BAET import app_console, configure_logging, create_logger
from BAET.app_args import get_args
from BAET.extract import MultiTrackAudioBulkExtractor

install(show_locals=True)


def main() -> None:
    args = get_args()

    if args.debug_options.print_args:
        rich.print(args)
        sys.exit(0)

    if args.debug_options.logging:
        log_path = Path("~/.baet").expanduser()
        log_path.mkdir(parents=True, exist_ok=True)
        log_file = log_path / f"logs_{datetime.now()}.txt"
        log_file.touch()

        configure_logging(enable_logging=True, file_out=log_file)
    else:
        configure_logging(enable_logging=False, file_out=None)

    logger = create_logger()
    extractor = MultiTrackAudioBulkExtractor(args)

    with Live(extractor, console=app_console):
        logger.info("Running jobs")
        extractor.run_synchronously()

    sys.exit(0)
